@@ -421,7 +421,14 @@ main(int argc, char **argv)
 	if (getconfs("langmap", sb))
 		langmap = check_strdup(strbuf_value(sb));
 	if (use_builtin_parser) {
-		parser_init(langmap, NULL);
+		const char *plugin_parser;
+
+		strbuf_reset(sb);
+		if (getconfs("gtags_parser", sb))
+			plugin_parser = strbuf_value(sb);
+		else
+			plugin_parser = NULL;
+		parser_init(langmap, plugin_parser);
 	} else {
 		set_env("GTAGSLANGMAP", langmap);
 		set_env("GTAGSDBPATH", dbpath);
@@ -795,23 +802,27 @@ updatetags(const char *dbpath, const char *root, IDSET *deleteset, STRBUF *addli
 	if (!getconfs(dbname(db), comline))
 		die("cannot get tag command. (%s)", dbname(db));
 	gtop = gtags_open(dbpath, root, db, GTAGS_MODIFY, 0);
-	if (vflag) {
-		char fid[32];
-		const char *path;
-		int total = idset_count(deleteset);
-		unsigned int id;
+	/*
+	 * Delete tags from GTAGS.
+	 */
+	if (!idset_empty(deleteset)) {
+		if (vflag) {
+			char fid[32];
+			const char *path;
+			int total = idset_count(deleteset);
+			unsigned int id;
 
-		seqno = 1;
-		for (id = idset_first(deleteset); id != END_OF_ID; id = idset_next(deleteset)) {
-			snprintf(fid, sizeof(fid), "%d", id);
-			path = gpath_fid2path(fid, NULL);
-			if (path == NULL)
-				die("GPATH is corrupted.");
-			fprintf(stderr, " [%d/%d] deleting tags of %s\n", seqno++, total, path + 2);
+			seqno = 1;
+			for (id = idset_first(deleteset); id != END_OF_ID; id = idset_next(deleteset)) {
+				snprintf(fid, sizeof(fid), "%d", id);
+				path = gpath_fid2path(fid, NULL);
+				if (path == NULL)
+					die("GPATH is corrupted.");
+				fprintf(stderr, " [%d/%d] deleting tags of %s\n", seqno++, total, path + 2);
+			}
 		}
-	}
-	if (!idset_empty(deleteset))
 		gtags_delete(gtop, deleteset);
+	}
 	gtop->flags = 0;
 	if (extractmethod)
 		gtop->flags |= GTAGS_EXTRACTMETHOD;
@@ -1029,22 +1040,26 @@ updatetags_using_builtin_parser(const char *dbpath, const char *root, IDSET *del
 	if (vflag)
 		fprintf(stderr, "[%s] Updating '%s'.\n", now(), dbname(GTAGS));
 	data.gtop[GTAGS] = gtags_open(dbpath, root, GTAGS, GTAGS_MODIFY, 0);
-	if (vflag) {
-		char fid[32];
-		int total = idset_count(deleteset);
-		unsigned int id;
+	/*
+	 * Delete tags from GTAGS.
+	 */
+	if (!idset_empty(deleteset)) {
+		if (vflag) {
+			char fid[32];
+			int total = idset_count(deleteset);
+			unsigned int id;
 
-		seqno = 1;
-		for (id = idset_first(deleteset); id != END_OF_ID; id = idset_next(deleteset)) {
-			snprintf(fid, sizeof(fid), "%d", id);
-			path = gpath_fid2path(fid, NULL);
-			if (path == NULL)
-				die("GPATH is corrupted.");
-			fprintf(stderr, " [%d/%d] deleting tags of %s\n", seqno++, total, path + 2);
+			seqno = 1;
+			for (id = idset_first(deleteset); id != END_OF_ID; id = idset_next(deleteset)) {
+				snprintf(fid, sizeof(fid), "%d", id);
+				path = gpath_fid2path(fid, NULL);
+				if (path == NULL)
+					die("GPATH is corrupted.");
+				fprintf(stderr, " [%d/%d] deleting tags of %s\n", seqno++, total, path + 2);
+			}
 		}
-	}
-	if (!idset_empty(deleteset))
 		gtags_delete(data.gtop[GTAGS], deleteset);
+	}
 	data.gtop[GTAGS]->flags = 0;
 	if (extractmethod)
 		data.gtop[GTAGS]->flags |= GTAGS_EXTRACTMETHOD;
@@ -1053,14 +1068,25 @@ updatetags_using_builtin_parser(const char *dbpath, const char *root, IDSET *del
 	if (test("f", makepath(dbpath, dbname(GRTAGS), NULL))) {
 		data.gtop[GRTAGS] = gtags_open(dbpath, root, GRTAGS, GTAGS_MODIFY, 0);
 		data.gtop[GRTAGS]->flags = data.gtop[GTAGS]->flags;
+		/*
+		 * If you set file pointer to tmpfile_fp, gtags_put() and gtags_put_using()
+		 * write records to the temporary file instead of db(3) file.
+		 */
 		tmp = tmpfile();
 		if (tmp == NULL)
 			die("cannot make temporary file.");
-		data.gtop[GRTAGS]->fp = tmp;
+		data.gtop[GRTAGS]->tmpfile_fp = tmp;
 	} else {
+		/*
+		 * If you set NULL to data.gtop[GRTAGS], parse_file() doesn't write to
+		 * GRTAGS. See put_syms().
+		 */
 		data.gtop[GRTAGS] = NULL;
 		tmp = NULL;
 	}
+	/*
+	 * Add tags to GTAGS and the temporary file.
+	 */
 	start = strbuf_value(addlist);
 	end = start + strbuf_getlen(addlist);
 	seqno = 0;
@@ -1076,6 +1102,7 @@ updatetags_using_builtin_parser(const char *dbpath, const char *root, IDSET *del
 		if (data.gtop[GRTAGS] != NULL)
 			gtags_flush(data.gtop[GRTAGS], data.fid);
 	}
+	parser_exit();
 	if (data.gtop[GRTAGS] == NULL) {
 		gtags_close(data.gtop[GTAGS]);
 		return;
@@ -1083,37 +1110,56 @@ updatetags_using_builtin_parser(const char *dbpath, const char *root, IDSET *del
 
 	if (vflag)
 		fprintf(stderr, "[%s] Updating '%s' and '%s'.\n", now(), dbname(GRTAGS), dbname(GSYMS));
-	data.gtop[GRTAGS]->fp = NULL;
+	data.gtop[GRTAGS]->tmpfile_fp = NULL;
 	data.gtop[GSYMS] = gtags_open(dbpath, root, GSYMS, GTAGS_MODIFY, 0);
-	if (vflag) {
-		char fid[32];
-		int total = idset_count(deleteset);
-		unsigned int id;
-
-		seqno = 1;
-		for (id = idset_first(deleteset); id != END_OF_ID; id = idset_next(deleteset)) {
-			snprintf(fid, sizeof(fid), "%d", id);
-			path = gpath_fid2path(fid, NULL);
-			if (path == NULL)
-				die("GPATH is corrupted.");
-			fprintf(stderr, " [%d/%d] deleting tags of %s\n", seqno++, total, path + 2);
-		}
-	}
+	/*
+	 * Delete tags from GRTAGS and GSYMS.
+	 */
 	if (!idset_empty(deleteset)) {
+		if (vflag) {
+			char fid[32];
+			int total = idset_count(deleteset);
+			unsigned int id;
+
+			seqno = 1;
+			for (id = idset_first(deleteset); id != END_OF_ID; id = idset_next(deleteset)) {
+				snprintf(fid, sizeof(fid), "%d", id);
+				path = gpath_fid2path(fid, NULL);
+				if (path == NULL)
+					die("GPATH is corrupted.");
+				fprintf(stderr, " [%d/%d] deleting tags of %s\n", seqno++, total, path + 2);
+			}
+		}
 		gtags_delete(data.gtop[GRTAGS], deleteset);
 		gtags_delete(data.gtop[GSYMS], deleteset);
 	}
+	/*
+	 * Move tags between GRTAGS and GSYMS.
+	 *
+	 * GRTAGS if not defined in GTAGS       ===> GSYMS
+	 * GSYMS if defined in GTAGS            ===> GRTAGS
+	 */
 	if (vflag) {
 		fprintf(stderr, " moving undefined symbols from '%s' to '%s'.\n", dbname(GRTAGS), dbname(GSYMS));
 		fprintf(stderr, " moving defined symbols from '%s' to '%s'.\n", dbname(GSYMS), dbname(GRTAGS));
 	}
 	gtags_move_ref_sym(data.gtop);
-	if (vflag) {
-		seqno = 0;
-		for (path = start; path < end; path += strlen(path) + 1)
-			fprintf(stderr, " [%d/%d] adding tags of %s\n", ++seqno, total, path + 2);
+	/*
+	 * Add tags to GRTAGS and GSYMS.
+	 *
+	 * temporary file ===>  if defined in GTAGS     ===> GRTAGS
+	 *                      else                    ===> GSYMS
+	 *
+	 * Gtags always makes GRTAGS and GSYMS even if they are empty.
+	 */
+	if (ftell(tmp) > 0) {
+		if (vflag) {
+			seqno = 0;
+			for (path = start; path < end; path += strlen(path) + 1)
+				fprintf(stderr, " [%d/%d] adding tags of %s\n", ++seqno, total, path + 2);
+		}
+		gtags_add_ref_sym(data.gtop, tmp);
 	}
-	gtags_add_ref_sym(data.gtop, tmp);
 	fclose(tmp);
 	gtags_close(data.gtop[GTAGS]);
 	gtags_close(data.gtop[GRTAGS]);
@@ -1150,7 +1196,14 @@ createtags_using_builtin_parser(const char *dbpath, const char *root)
 		data.gtop[GTAGS]->flags |= GTAGS_DEBUG;
 	data.gtop[GRTAGS] = gtags_open(dbpath, root, GRTAGS, GTAGS_CREATE, openflags);
 	data.gtop[GRTAGS]->flags = data.gtop[GTAGS]->flags;
-	data.gtop[GRTAGS]->fp = tmp;
+	/*
+	 * If you set file pointer to tmpfile_fp, gtags_put() and gtags_put_using()
+	 * write records to the temporary file instead of db(3) file.
+	 */
+	data.gtop[GRTAGS]->tmpfile_fp = tmp;
+	/*
+	 * Add tags to GTAGS and the temporary file.
+	 */
 	if (file_list)
 		find_open_filelist(file_list, root);
 	else
@@ -1175,8 +1228,9 @@ createtags_using_builtin_parser(const char *dbpath, const char *root)
 		gtags_flush(data.gtop[GRTAGS], data.fid);
 	}
 	total = seqno;
+	parser_exit();
 	find_close();
-	data.gtop[GRTAGS]->fp = NULL;
+	data.gtop[GRTAGS]->tmpfile_fp = NULL;
 	statistics_time_end(tim);
 	strbuf_reset(sb);
 	if (getconfs("GTAGS_extra", sb)) {
@@ -1198,20 +1252,30 @@ createtags_using_builtin_parser(const char *dbpath, const char *root)
 	if (vflag)
 		fprintf(stderr, "[%s] Creating '%s' and '%s'.\n", now(), dbname(GRTAGS), dbname(GSYMS));
 	data.gtop[GSYMS] = gtags_open(dbpath, root, GSYMS, GTAGS_CREATE, openflags);
-	if (vflag) {
-		if (file_list)
-			find_open_filelist(file_list, root);
-		else
-			find_open(NULL);
-		seqno = 0;
-		while ((path = find_read()) != NULL) {
-			if (*path == ' ')
-				continue;
-			fprintf(stderr, " [%d/%d] adding tags of %s\n", ++seqno, total, path + 2);
+	/*
+	 * Add tags to GRTAGS and GSYMS.
+	 *
+	 * temporary file ===>  if defined in GTAGS     ===> GRTAGS
+	 *                      else                    ===> GSYMS
+	 *
+	 * Gtags always makes GRTAGS and GSYMS even if they are empty.
+	 */
+	if (ftell(tmp) > 0) {
+		if (vflag) {
+			if (file_list)
+				find_open_filelist(file_list, root);
+			else
+				find_open(NULL);
+			seqno = 0;
+			while ((path = find_read()) != NULL) {
+				if (*path == ' ')
+					continue;
+				fprintf(stderr, " [%d/%d] adding tags of %s\n", ++seqno, total, path + 2);
+			}
+			find_close();
 		}
-		find_close();
+		gtags_add_ref_sym(data.gtop, tmp);
 	}
-	gtags_add_ref_sym(data.gtop, tmp);
 	fclose(tmp);
 	statistics_time_end(tim);
 	tim = statistics_time_start("Time of flushing B-tree cache");
