@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
- *      2006, 2007, 2008 Tama Communications Corporation
+ *      2006, 2007, 2008, 2010 Tama Communications Corporation
  *
  * This file is part of GNU GLOBAL.
  *
@@ -37,7 +37,6 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/param.h>
-#include <time.h>
 
 #include "checkalloc.h"
 #include "getopt.h"
@@ -56,6 +55,7 @@ int makedupindex(void);
 int makedefineindex(const char *, int, STRBUF *);
 int makefileindex(const char *, STRBUF *);
 void makeincludeindex(void);
+int makecflowindex(const char *, const char *);
 
 #if defined(_WIN32) && !defined(__CYGWIN__)
 #define mkdir(path,mode) mkdir(path)
@@ -92,8 +92,9 @@ const char *tmpdir = "/tmp";
  * 'm': mains
  * 'd': definitions
  * 'f': files
+ * 't': call tree
  */
-char *item_order = "csmdf";
+char *item_order = "csmdft";
 /*
  * options
  */
@@ -116,7 +117,7 @@ int caution;				/* --caution option		*/
 int dynamic;				/* --dynamic(-D) option		*/
 int symbol;				/* --symbol(-s) option          */
 int suggest;				/* --suggest option		*/
-int statistics;				/* --statistics option		*/
+int statistics = STATISTICS_STYLE_NONE;	/* --statistics option		*/
 
 int copy_files;				/* 1: copy tag files		*/
 int no_order_list;			/* 1: doesn't use order list	*/
@@ -143,6 +144,7 @@ const char *insert_footer;		/* --insert-footer=<file>	*/
  */
 const char *title_define_index = "DEFINITIONS";
 const char *title_file_index = "FILES";
+const char *title_call_tree = "CALL TREE";
 const char *title_included_from = "INCLUDED FROM";
 /*
  * Function header items.
@@ -225,6 +227,7 @@ const char *icon_suffix = "png";	/* icon suffix (jpg, png etc)	*/
 const char *icon_spec = "border='0' align='top'";/* parameter in IMG tag*/
 const char *prolog_script = NULL;	/* include script at first	*/
 const char *epilog_script = NULL;	/* include script at last	*/
+const char *cflow_file = NULL;		/* file name of cflow output	*/
 int show_position = 0;			/* show current position	*/
 int table_list = 0;			/* tag list using table tag	*/
 int table_flist = 0;			/* file list using table tag	*/
@@ -278,7 +281,7 @@ static struct option const long_options[] = {
         {"nocgi", no_argument, &cgi, 0},
         {"no-map-file", no_argument, &map_file, 0},
         {"show-position", no_argument, &show_position, 1},
-        {"statistics", no_argument, &statistics, 1},
+        {"statistics", no_argument, &statistics, STATISTICS_STYLE_TABLE},
         {"suggest", no_argument, &suggest, 1},
         {"table-list", no_argument, &table_list, 1},
         {"version", no_argument, &show_version, 1},
@@ -296,7 +299,9 @@ static struct option const long_options[] = {
 #define OPT_INSERT_HEADER	136
 #define OPT_ITEM_ORDER		137
 #define OPT_TABS		138
+#define OPT_CFLOW		139
         {"action", required_argument, NULL, OPT_ACTION},
+        {"cflow", required_argument, NULL, OPT_CFLOW},
         {"cvsweb", required_argument, NULL, OPT_CVSWEB},
         {"cvsweb-cvsroot", required_argument, NULL, OPT_CVSWEB_CVSROOT},
         {"gtagsconf", required_argument, NULL, OPT_GTAGSCONF},
@@ -954,6 +959,16 @@ makecommonpart(const char *title, const char *defines, const char *files)
 				strbuf_puts_nl(sb, hr);
 			}
 			break;
+		case 't':
+			if (cflow_file) {
+				strbuf_puts(sb, header_begin);
+				strbuf_puts(sb, gen_href_begin(NULL, "cflow", normal_suffix, NULL));
+				strbuf_puts(sb, title_call_tree);
+				strbuf_puts(sb, gen_href_end());
+				strbuf_puts_nl(sb, header_end);
+				strbuf_puts_nl(sb, hr);
+			}
+			break;
 		case 'm':
 			strbuf_sprintf(sb, "%sMAINS%s\n", header_begin, header_end);
 
@@ -1007,8 +1022,8 @@ makecommonpart(const char *title, const char *defines, const char *files)
 					strbuf_puts_nl(sb, list_end);
 				else
 					strbuf_puts_nl(sb, br);
-				strbuf_puts_nl(sb, hr);
 			}
+			strbuf_puts_nl(sb, hr);
 			break;
 		default:
 			warning("unknown item '%c'. (Ignored)", *item);
@@ -1049,7 +1064,7 @@ basic_check(void)
  * load configuration variables.
  */
 static void
-configuration(int argc, char **argv)
+configuration(int argc, char *const *argv)
 {
 	STRBUF *sb = strbuf_open(0);
 	int i, n;
@@ -1313,7 +1328,7 @@ configuration(int argc, char **argv)
  * save_environment: save configuration data and arguments.
  */
 static void
-save_environment(int argc, char **argv)
+save_environment(int argc, char *const *argv)
 {
 	char command[MAXFILLEN];
 	STRBUF *sb = strbuf_open(0);
@@ -1381,7 +1396,7 @@ save_environment(int argc, char **argv)
 }
 
 char **
-append_options(int *argc, char **argv)
+append_options(int *argc, char *const *argv)
 {
 
 	STRBUF *sb = strbuf_open(0);
@@ -1427,15 +1442,14 @@ append_options(int *argc, char **argv)
 	while (j < *argc)
 		newargv[i++] = argv[j++];
 	newargv[i] = NULL;
-	argv = (char **)newargv;
 	*argc = i;
 #ifdef DEBUG
 	for (i = 0; i < *argc; i++)
-		fprintf(stderr, "argv[%d] = '%s'\n", i, argv[i]);
+		fprintf(stderr, "newargv[%d] = '%s'\n", i, newargv[i]);
 #endif
 	/* doesn't close string buffer. */
 
-	return argv;
+	return (char **)newargv;
 }
 int
 main(int argc, char **argv)
@@ -1446,9 +1460,7 @@ main(int argc, char **argv)
 	const char *index = NULL;
 	int optchar;
         int option_index = 0;
-	time_t start_time, end_time, start_all_time, end_all_time,
-		T_makedupindex, T_makedefineindex, T_makefileindex,
-		T_makeincludeindex, T_makehtml, T_all;
+	STATISTICS_TIME *tim;
 
 	arg_dbpath[0] = 0;
 	basic_check();
@@ -1470,6 +1482,9 @@ main(int argc, char **argv)
 		case OPT_ACTION:
 			action_value = optarg;
 			break;
+		case OPT_CFLOW:
+			cflow_file = optarg;
+			break;
 		case OPT_CVSWEB:
 			cvsweb_url = optarg;
 			break;
@@ -1616,6 +1631,8 @@ main(int argc, char **argv)
 		if (gtags_not_found)
 			gflag = 1;
 	}
+	if (cflow_file && !test("fr", cflow_file))
+		die("cflow file not found. '%s'", cflow_file);
 	if (insert_header && !test("fr", insert_header))
 		die("page header file '%s' not found.", insert_header);
 	if (insert_footer && !test("fr", insert_footer))
@@ -1820,7 +1837,7 @@ main(int argc, char **argv)
         HTML = (cflag) ? gzipped_suffix : normal_suffix;
 
 	message("[%s] Htags started", now());
-	start_all_time = time(NULL);
+	init_statistics();
 	/*
 	 * (#) check if GTAGS, GRTAGS is the latest.
 	 */
@@ -1895,11 +1912,10 @@ main(int argc, char **argv)
 	 */
 	message("[%s] (3) making duplicate entries ...", now());
 	cache_open();
-	start_time = time(NULL);
+	tim = statistics_time_start("Time of making duplicate entries");
 	func_total = makedupindex();
-	end_time = time(NULL);
+	statistics_time_end(tim);
 	message("Total %d functions.", func_total);
-	T_makedupindex = end_time - start_time;
 	/*
 	 * (4) search index. (search.html)
 	 */
@@ -1916,31 +1932,38 @@ main(int argc, char **argv)
 		 *     PRODUCE @defines
 		 */
 		message("[%s] (5) making function index ...", now());
-		start_time = time(NULL);
+		tim = statistics_time_start("Time of making function index");
 		func_total = makedefineindex("defines.html", func_total, defines);
-		end_time = time(NULL);
+		statistics_time_end(tim);
 		message("Total %d functions.", func_total);
-		T_makedefineindex = end_time - start_time;
 		/*
 		 * (6) make file index (files.html and files/)
 		 *     PRODUCE @files, %includes
 		 */
 		message("[%s] (6) making file index ...", now());
 		init_inc();
-		start_time = time(NULL);
+		tim = statistics_time_start("Time of making file index");
 		file_total = makefileindex("files.html", files);
-		end_time = time(NULL);
+		statistics_time_end(tim);
 		message("Total %d files.", file_total);
-		T_makefileindex = end_time - start_time;
 		html_count += file_total;
+		/*
+		 * (7) make call tree using cflow(1)'s output (cflow.html)
+		 */
+		if (cflow_file) {
+			message("[%s] (7) making cflow index ...", now());
+			tim = statistics_time_start("Time of making cflow index");
+			if (makecflowindex("cflow.html", cflow_file) < 0)
+				cflow_file = NULL;
+			statistics_time_end(tim);
+		}
 		/*
 		 * [#] make include file index.
 		 */
 		message("[%s] (#) making include file index ...", now());
-		start_time = time(NULL);
+		tim = statistics_time_start("Time of making include file index");
 		makeincludeindex();
-		end_time = time(NULL);
-		T_makeincludeindex = end_time - start_time;
+		statistics_time_end(tim);
 		/*
 		 * [#] make a common part for mains.html and index.html
 		 *     USING @defines @files
@@ -1966,10 +1989,9 @@ main(int argc, char **argv)
 	 *     USING TAG CACHE, %includes and anchor database.
 	 */
 	message("[%s] (9) making hypertext from source code ...", now());
-	start_time = time(NULL);
+	tim = statistics_time_start("Time of making hypertext");
 	makehtml(file_total);
-	end_time = time(NULL);
-	T_makehtml = end_time - start_time;
+	statistics_time_end(tim);
 	/*
 	 * (10) rebuild script. (rebuild.sh)
 	 *
@@ -1987,9 +2009,6 @@ main(int argc, char **argv)
 		snprintf(dst, sizeof(dst), "%s/style.css", distpath);
 		copyfile(src, dst);
 	}
-	end_all_time = time(NULL);
-	message("[%s] Done.", now());
-	T_all = end_all_time - start_all_time;
 	if (vflag && cgi && (cflag || fflag || dynamic)) {
 		message("\n[Information]\n");
 		if (cflag) {
@@ -2025,15 +2044,7 @@ main(int argc, char **argv)
 	/*
 	 * Print statistics information.
 	 */
-	if (statistics) {
-		setverbose();
-		message("- Elapsed time of making duplicate entries ............ %10ld seconds.", T_makedupindex);
-		message("- Elapsed time of making function index ............... %10ld seconds.", T_makedefineindex);
-		message("- Elapsed time of making file index ................... %10ld seconds.", T_makefileindex);
-		message("- Elapsed time of making include file index ........... %10ld seconds.", T_makeincludeindex);
-		message("- Elapsed time of making hypertext .................... %10ld seconds.", T_makehtml);
-		message("- The entire elapsed time ............................. %10ld seconds.", T_all);
-	}
+	print_statistics(statistics);
 	clean();
 	return 0;
 }
