@@ -334,13 +334,13 @@ openconf(void)
 #endif /* _WIN32 */
 		strbuf_puts(sb, ":GTAGS=");
 		strbuf_puts(sb, path);
-		strbuf_puts(sb, " -dt %s");
+		strbuf_puts(sb, " %s");
 		strbuf_puts(sb, ":GRTAGS=");
 		strbuf_puts(sb, path);
-		strbuf_puts(sb, " -dtr %s");
+		strbuf_puts(sb, " -r %s");
 		strbuf_puts(sb, ":GSYMS=");
 		strbuf_puts(sb, path);
-		strbuf_puts(sb, " -dts %s");
+		strbuf_puts(sb, " -s %s");
 	}
 	strbuf_unputc(sb, ':');
 	strbuf_putc(sb, ':');
@@ -390,7 +390,8 @@ getconfs(const char *name, STRBUF *sb)
 
 	if (!opened)
 		openconf();
-	if (!strcmp(name, "suffixes") || !strcmp(name, "skip"))
+	if (!strcmp(name, "suffixes") || !strcmp(name, "skip")
+	 || !strcmp(name, "gtags_parser") || !strcmp(name, "langmap"))
 		all = 1;
 	snprintf(buf, sizeof(buf), ":%s=", name);
 	p = confline;
