@@ -121,6 +121,26 @@ seekto(const char *string, int n)
 	}
 	return p;
 }
+/*
+ * tmpfile_put: write to the temporary file instead of db(3) file.
+ *
+ *	i)	fp	file pointer
+ *	i)	key	key of db(3)
+ *	i)	data	data of db(3)
+ *
+ * record format:
+ * <key>\t<data>\n
+ *
+ * Finally, this record should be written in db(3) file.
+ */
+void
+tmpfile_put(FILE *fp, const char *key, const char *data)
+{
+	fputs(key, fp);
+	putc('\t', fp);
+	fputs(data, fp);
+	putc('\n', fp);
+}
 /*
  * Tag format
  *
@@ -431,23 +451,17 @@ gtags_put_using(GTOP *gtop, const char *tag, int lno, const char *fid, const cha
 	} else {
 		key = tag;
 	}
-	if (gtop->fp != NULL) {
-		fputs(key, gtop->fp);
-		putc('\n', gtop->fp);
-		fputs(fid, gtop->fp);
-		putc(' ', gtop->fp);
-		fputs((gtop->format & GTAGS_COMPNAME) ? compress(tag, key) : tag, gtop->fp);
-		fprintf(gtop->fp, " %d %s\n", lno,
-			(gtop->format & GTAGS_COMPRESS) ? compress(img, key) : img);
+	strbuf_reset(gtop->sb);
+	strbuf_puts(gtop->sb, fid);
+	strbuf_putc(gtop->sb, ' ');
+	strbuf_puts(gtop->sb, (gtop->format & GTAGS_COMPNAME) ? compress(tag, key) : tag);
+	strbuf_putc(gtop->sb, ' ');
+	strbuf_putn(gtop->sb, lno);
+	strbuf_putc(gtop->sb, ' ');
+	strbuf_puts(gtop->sb, (gtop->format & GTAGS_COMPRESS) ? compress(img, key) : img);
+	if (gtop->tmpfile_fp != NULL) {
+		tmpfile_put(gtop->tmpfile_fp, key, strbuf_value(gtop->sb));
 	} else {
-		strbuf_reset(gtop->sb);
-		strbuf_puts(gtop->sb, fid);
-		strbuf_putc(gtop->sb, ' ');
-		strbuf_puts(gtop->sb, (gtop->format & GTAGS_COMPNAME) ? compress(tag, key) : tag);
-		strbuf_putc(gtop->sb, ' ');
-		strbuf_putn(gtop->sb, lno);
-		strbuf_putc(gtop->sb, ' ');
-		strbuf_puts(gtop->sb, (gtop->format & GTAGS_COMPRESS) ? compress(img, key) : img);
 		dbop_put(gtop->dbop, key, strbuf_value(gtop->sb));
 	}
 }
@@ -476,24 +490,27 @@ void
 gtags_add_ref_sym(GTOP *const *gtop, FILE *ip)
 {
 	STRBUF *ib = strbuf_open(MAXBUFLEN);
-	int data_offset;
-	const char *key;
-	DBOP *dbop[GTAGLIM];
 
-	dbop[GTAGS] = gtop[GTAGS]->dbop;
-	dbop[GRTAGS] = gtop[GRTAGS]->dbop;
-	dbop[GSYMS] = gtop[GSYMS]->dbop;
 	rewind(ip);
 	while (strbuf_fgets(ib, ip, STRBUF_NOCRLF) != NULL) {
-		strbuf_putc(ib, '\0');
-		data_offset = strbuf_getlen(ib);
-		key = strbuf_fgets(ib, ip, STRBUF_NOCRLF | STRBUF_APPEND);
-		if (dbop_get(dbop[GTAGS], key) != NULL)
-			dbop_put(dbop[GRTAGS], key, key + data_offset);
+		/*
+		 * [record format]
+		 * <key>\t<data>\n
+		 */
+		char *key = strbuf_value(ib);
+		char *data = strchr(key, '\t');
+		if (data == NULL)
+			die("gtags_add_ref_sym: internal error.");
+		*data++ = '\0';
+		/*
+		 * If the key is defined in GTAGS then put the record into GRTAGS
+		 * else put it into GSYMS.
+		 */
+		if (dbop_get(gtop[GTAGS]->dbop, key) != NULL)
+			dbop_put(gtop[GRTAGS]->dbop, key, data);
 		else
-			dbop_put(dbop[GSYMS], key, key + data_offset);
+			dbop_put(gtop[GSYMS]->dbop, key, data);
 	}
-
 	strbuf_close(ib);
 }
 /*
@@ -506,37 +523,58 @@ void
 gtags_move_ref_sym(GTOP *const *gtop)
 {
 	int result = 0;
-	const char *defkey, *refdata, *symdata;
-	DBOP *dbop[GTAGLIM];
+	DBOP *defdbop;
+	const char *defkey;
+	struct {
+		DBOP *dbop;
+		const char *key, *dat;
+		char saved_key[MAXKEYLEN+1];
+		STRBUF *saved_dat;
+	} ref, sym, *smaller, *larger;
 
-	dbop[GTAGS] = gtop[GTAGS]->dbop;
-	dbop[GRTAGS] = gtop[GRTAGS]->dbop;
-	dbop[GSYMS] = gtop[GSYMS]->dbop;
-	defkey = dbop_first(dbop[GTAGS], NULL, NULL, DBOP_KEY);
-	refdata = dbop_first(dbop[GRTAGS], NULL, NULL, 0);
-	symdata = dbop_first(dbop[GSYMS], NULL, NULL, 0);
+	defdbop = gtop[GTAGS]->dbop;
+	defkey = dbop_first(defdbop, NULL, NULL, DBOP_KEY);
+	ref.dbop = gtop[GRTAGS]->dbop;
+	ref.dat = dbop_first(ref.dbop, NULL, NULL, 0);
+	ref.key = ref.dbop->lastkey;
+	ref.saved_dat = strbuf_open(MAXBUFLEN);
+	sym.dbop = gtop[GSYMS]->dbop;
+	sym.dat = dbop_first(sym.dbop, NULL, NULL, 0);
+	sym.key = sym.dbop->lastkey;
+	sym.saved_dat = strbuf_open(MAXBUFLEN);
 	for (;;) {
-		if (refdata != NULL
-		    && (symdata == NULL || strcmp(dbop[GRTAGS]->lastkey, dbop[GSYMS]->lastkey) < 0)) {
-			while (defkey != NULL && (result = strcmp(dbop[GRTAGS]->lastkey, defkey)) > 0)
-				defkey = dbop_next(dbop[GTAGS]);
-			if (defkey == NULL || result < 0) {
-				dbop_put(dbop[GSYMS], dbop[GRTAGS]->lastkey, refdata);
-				dbop_delete(dbop[GRTAGS], NULL);
-			}
-			refdata = dbop_next(dbop[GRTAGS]);
-		} else if (symdata != NULL) {
-			while (defkey != NULL && (result = strcmp(dbop[GSYMS]->lastkey, defkey)) > 0)
-				defkey = dbop_next(dbop[GTAGS]);
-			if (defkey != NULL && result == 0) {
-				dbop_put(dbop[GRTAGS], dbop[GSYMS]->lastkey, symdata);
-				dbop_delete(dbop[GSYMS], NULL);
-			}
-			symdata = dbop_next(dbop[GSYMS]);
+		if (ref.dat != NULL && (sym.dat == NULL || strcmp(ref.key, sym.key) < 0)) {
+			smaller = &ref;
+			larger = &sym;
+		} else if (sym.dat != NULL) {
+			smaller = &sym;
+			larger = &ref;
 		} else {
 			break;
 		}
+		while (defkey != NULL && (result = strcmp(smaller->key, defkey)) > 0)
+			defkey = dbop_next(defdbop);
+		if (defkey != NULL && (smaller == &ref ? result < 0 : result == 0)) {
+			/*
+			 * Calling dbop_put doesn't affect the position of cursor,
+			 * but it invalidates the pointer to read data.
+			 */
+			if (larger->dat != NULL && larger->key != larger->saved_key) {
+				strlimcpy(larger->saved_key, larger->key, sizeof(larger->saved_key));
+				larger->key = larger->saved_key;
+				strbuf_reset(larger->saved_dat);
+				strbuf_puts(larger->saved_dat, larger->dat);
+				larger->dat = strbuf_value(larger->saved_dat);
+			}
+			dbop_put(larger->dbop, smaller->key, smaller->dat);
+			dbop_delete(smaller->dbop, NULL);
+		}
+		smaller->dat = dbop_next(smaller->dbop);
+		smaller->key = smaller->dbop->lastkey;
 	}
+
+	strbuf_close(ref.saved_dat);
+	strbuf_close(sym.saved_dat);
 }
 /*
  * gtags_put: put tag record with packing.
@@ -605,7 +643,11 @@ gtags_put(GTOP *gtop, const char *key, const char *ctags_x)	/* virtually const *
 		strbuf_puts(gtop->sb, gtop->format & GTAGS_COMPRESS ?
 			compress(ptable.part[PART_LINE].start, key) :
 			ptable.part[PART_LINE].start);
-		dbop_put(gtop->dbop, key, strbuf_value(gtop->sb));
+		if (gtop->tmpfile_fp != NULL) {
+			tmpfile_put(gtop->tmpfile_fp, key, strbuf_value(gtop->sb));
+		} else {
+			dbop_put(gtop->dbop, key, strbuf_value(gtop->sb));
+		}
 	}
 	recover(&ptable);
 }
@@ -959,11 +1001,8 @@ flush_pool(GTOP *gtop, const char *s_fid)
 						strbuf_putn(gtop->sb, n);
 					}
 					if (strbuf_getlen(gtop->sb) > DBOP_PAGESIZE / 4) {
-						if (gtop->fp != NULL) {
-							fputs(key, gtop->fp);
-							putc('\n', gtop->fp);
-							fputs(strbuf_value(gtop->sb), gtop->fp);
-							putc('\n', gtop->fp);
+						if (gtop->tmpfile_fp != NULL) {
+							tmpfile_put(gtop->tmpfile_fp, key, strbuf_value(gtop->sb));
 						} else {
 							dbop_put(gtop->dbop, key, strbuf_value(gtop->sb));
 						}
@@ -990,11 +1029,8 @@ flush_pool(GTOP *gtop, const char *s_fid)
 					strbuf_putc(gtop->sb, ',');
 				strbuf_putn(gtop->sb, n);
 				if (strbuf_getlen(gtop->sb) > DBOP_PAGESIZE / 4) {
-					if (gtop->fp != NULL) {
-						fputs(key, gtop->fp);
-						putc('\n', gtop->fp);
-						fputs(strbuf_value(gtop->sb), gtop->fp);
-						putc('\n', gtop->fp);
+					if (gtop->tmpfile_fp != NULL) {
+						tmpfile_put(gtop->tmpfile_fp, key, strbuf_value(gtop->sb));
 					} else {
 						dbop_put(gtop->dbop, key, strbuf_value(gtop->sb));
 					}
@@ -1004,11 +1040,8 @@ flush_pool(GTOP *gtop, const char *s_fid)
 			}
 		}
 		if (strbuf_getlen(gtop->sb) > header_offset) {
-			if (gtop->fp != NULL) {
-				fputs(key, gtop->fp);
-				putc('\n', gtop->fp);
-				fputs(strbuf_value(gtop->sb), gtop->fp);
-				putc('\n', gtop->fp);
+			if (gtop->tmpfile_fp != NULL) {
+				tmpfile_put(gtop->tmpfile_fp, key, strbuf_value(gtop->sb));
 			} else {
 				dbop_put(gtop->dbop, key, strbuf_value(gtop->sb));
 			}
